button.active-menu {
  color: #2e3436;
  outline-color: rgba(46, 52, 54, 0.3);
  border-color: #b6b6b3;
  background-image: image(#d9d9d7);
  box-shadow: inset 0 1px rgba(255, 255, 255, 0);
  text-shadow: none;
  -gtk-icon-shadow: none; }

.floating-bar {
  background-color: @theme_base_color;
  border-width: 1px;
  border-style: solid solid none;
  border-color: @borders;
  border-radius: 3px 3px 0 0; }
  .floating-bar.left {
    border-left-style: none;
    border-top-left-radius: 0; }
  .floating-bar.right {
    border-right-style: none;
    border-top-right-radius: 0; }
  .floating-bar:backdrop {
    background-color: @theme_unfocused_base_color;
    border-color: @unfocused_borders; }

.incognito-mode.titlebar {
  background: #cbd2d9 -gtk-icontheme("user-not-tracked-symbolic") 160px 0/64px 64px no-repeat, linear-gradient(to top, #a5b1bd, #c5cdd5 2px, #cbd2d9 3px);
  box-shadow: inset 0 1px #f1f3f5;
  border-color: #909fae;
  color: rgba(46, 52, 54, 0.2); }
  .incognito-mode.titlebar > * {
    color: #2e3436; }
    .incognito-mode.titlebar > *:backdrop {
      color: #8b8e8f; }
  .incognito-mode.titlebar:backdrop {
    background-image: -gtk-icontheme("user-not-tracked-symbolic"), image(#cbd2d9);
    box-shadow: inset 0 1px #f1f3f5;
    color: rgba(139, 142, 143, 0.1); }
  .incognito-mode.titlebar button {
    color: #2e3436;
    outline-color: rgba(46, 52, 54, 0.3);
    border-color: #909fae;
    border-bottom-color: #738698;
    background-image: linear-gradient(to bottom, #cbd2d9, #bfc8d0 60%, #aeb9c3);
    text-shadow: 0 1px rgba(255, 255, 255, 0.76923);
    -gtk-icon-shadow: 0 1px rgba(255, 255, 255, 0.76923);
    box-shadow: inset 0 1px rgba(255, 255, 255, 0.6); }
    .incognito-mode.titlebar button.flat, .incognito-mode.titlebar button.titlebutton {
      border-color: transparent;
      background-color: transparent;
      background-image: none;
      box-shadow: inset 0 1px rgba(255, 255, 255, 0);
      text-shadow: none;
      -gtk-icon-shadow: none; }
    .incognito-mode.titlebar button.titlebutton {
      text-shadow: 0 1px rgba(255, 255, 255, 0.76923);
      -gtk-icon-shadow: 0 1px rgba(255, 255, 255, 0.76923); }
    .incognito-mode.titlebar button:hover {
      color: #2e3436;
      outline-color: rgba(46, 52, 54, 0.3);
      border-color: #909fae;
      border-bottom-color: #738698;
      text-shadow: 0 1px rgba(255, 255, 255, 0.76923);
      -gtk-icon-shadow: 0 1px rgba(255, 255, 255, 0.76923);
      box-shadow: inset 0 1px rgba(255, 255, 255, 0.6);
      background-image: linear-gradient(to bottom, #dde1e6, #cbd2d9 60%, #bfc8d0); }
    .incognito-mode.titlebar button:active, .incognito-mode.titlebar button:checked, .incognito-mode.titlebar button#ephy-page-menu-button.active-menu {
      color: #2e3436;
      outline-color: rgba(46, 52, 54, 0.3);
      border-color: #909fae;
      background-image: image(#b9c3cc);
      box-shadow: inset 0 1px rgba(255, 255, 255, 0);
      text-shadow: none;
      -gtk-icon-shadow: none; }
    .incognito-mode.titlebar button:disabled {
      border-color: #909fae;
      background-image: image(#d3d9df);
      text-shadow: none;
      -gtk-icon-shadow: none;
      box-shadow: inset 0 1px rgba(255, 255, 255, 0); }
      .incognito-mode.titlebar button:disabled label, .incognito-mode.titlebar button:disabled {
        color: #8b8e8f; }
    .incognito-mode.titlebar button:backdrop {
      border-color: #cbd2d9;
      background-image: image(#cbd2d9);
      text-shadow: none;
      -gtk-icon-shadow: none;
      box-shadow: inset 0 1px rgba(255, 255, 255, 0);
      border-color: #909fae; }
      .incognito-mode.titlebar button:backdrop label, .incognito-mode.titlebar button:backdrop {
        color: #8b8e8f; }
      .incognito-mode.titlebar button:backdrop:active {
        border-color: #bbc5ce;
        background-image: image(#bbc5ce);
        box-shadow: inset 0 1px rgba(255, 255, 255, 0);
        border-color: #909fae; }
        .incognito-mode.titlebar button:backdrop:active label, .incognito-mode.titlebar button:backdrop:active {
          color: #8b8e8f; }
      .incognito-mode.titlebar button:backdrop:disabled {
        border-color: #d3d9df;
        background-image: image(#d3d9df);
        text-shadow: none;
        -gtk-icon-shadow: none;
        box-shadow: inset 0 1px rgba(255, 255, 255, 0);
        border-color: #909fae; }
        .incognito-mode.titlebar button:backdrop:disabled label, .incognito-mode.titlebar button:backdrop:disabled {
          color: #999fa4; }
    .incognito-mode.titlebar button.flat:backdrop, .incognito-mode.titlebar button.titlebutton:backdrop {
      border-color: transparent;
      background-color: transparent;
      background-image: none;
      box-shadow: inset 0 1px rgba(255, 255, 255, 0);
      text-shadow: none;
      -gtk-icon-shadow: none; }
  .incognito-mode.titlebar entry {
    border-color: #909fae; }
    .incognito-mode.titlebar entry:focus {
      border-color: #4a90d9;
      box-shadow: inset 0 0 0 1px #4a90d9; }
    .incognito-mode.titlebar entry:backdrop {
      box-shadow: none; }

/* Bookmarks */
.titlebar entry.starred image.right {
  color: #fce94f; }

.bookmarks-row button {
  opacity: 0.25; }
.bookmarks-row:hover button {
  opacity: 1; }

.bookmark-tag-widget {
  padding-left: 8px;
  background-color: #cfcfcd; }
  .bookmark-tag-widget label {
    padding-left: 8px;
    padding-right: 8px; }
  .bookmark-tag-widget button {
    background: none;
    border: none;
    box-shadow: none; }
    .bookmark-tag-widget button:last-child image {
      color: #454f52; }
    .bookmark-tag-widget button:last-child:hover image {
      border: 1px solid;
      border-radius: 3px;
      color: #454f52;
      outline-color: rgba(69, 79, 82, 0.3);
      border-color: #b6b6b3;
      border-bottom-color: #91918c;
      background-image: linear-gradient(to bottom, #e8e8e7, #dededd 60%, #cfcfcd);
      text-shadow: 0 1px rgba(255, 255, 255, 0.76923);
      -gtk-icon-shadow: 0 1px rgba(255, 255, 255, 0.76923);
      box-shadow: inset 0 1px rgba(255, 255, 255, 0.8);
      color: #454f52;
      outline-color: rgba(69, 79, 82, 0.3);
      border-color: #b6b6b3;
      border-bottom-color: #91918c;
      text-shadow: 0 1px rgba(255, 255, 255, 0.76923);
      -gtk-icon-shadow: 0 1px rgba(255, 255, 255, 0.76923);
      box-shadow: inset 0 1px white;
      background-image: linear-gradient(to bottom, #f7f7f7, #e8e8e7 60%, #dededd); }
    .bookmark-tag-widget button:last-child:active image {
      color: #454f52;
      outline-color: rgba(69, 79, 82, 0.3);
      border-color: #b6b6b3;
      border-bottom-color: #91918c;
      background-image: linear-gradient(to bottom, #e8e8e7, #dededd 60%, #cfcfcd);
      text-shadow: 0 1px rgba(255, 255, 255, 0.76923);
      -gtk-icon-shadow: 0 1px rgba(255, 255, 255, 0.76923);
      box-shadow: inset 0 1px rgba(255, 255, 255, 0.8);
      color: #454f52;
      outline-color: rgba(69, 79, 82, 0.3);
      border-color: #b6b6b3;
      background-image: image(#d9d9d7);
      box-shadow: inset 0 1px rgba(255, 255, 255, 0);
      text-shadow: none;
      -gtk-icon-shadow: none; }

.bookmark-tag-widget-selected {
  background-color: #4a90d9; }
  .bookmark-tag-widget-selected image {
    color: white; }
  .bookmark-tag-widget-selected label {
    color: white; }
  .bookmark-tag-widget-selected button:last-child image {
    color: white; }
  .bookmark-tag-widget-selected button:last-child:hover image {
    color: white;
    outline-color: rgba(255, 255, 255, 0.3);
    border-color: #215d9c;
    border-bottom-color: #184472;
    background-image: linear-gradient(to bottom, #4a90d9, #3986d5 60%, #2a76c6);
    text-shadow: 0 -1px rgba(0, 0, 0, 0.54353);
    -gtk-icon-shadow: 0 -1px rgba(0, 0, 0, 0.54353);
    box-shadow: inset 0 1px rgba(255, 255, 255, 0.4);
    color: white;
    outline-color: rgba(255, 255, 255, 0.3);
    border-color: #215d9c;
    border-bottom-color: #184472;
    text-shadow: 0 -1px rgba(0, 0, 0, 0.49553);
    -gtk-icon-shadow: 0 -1px rgba(0, 0, 0, 0.49553);
    box-shadow: inset 0 1px rgba(255, 255, 255, 0.4);
    background-image: linear-gradient(to bottom, #63a0de, #4a90d9 60%, #3986d5); }
  .bookmark-tag-widget-selected button:last-child:active image {
    color: white;
    outline-color: rgba(255, 255, 255, 0.3);
    border-color: #215d9c;
    border-bottom-color: #184472;
    background-image: linear-gradient(to bottom, #4a90d9, #3986d5 60%, #2a76c6);
    text-shadow: 0 -1px rgba(0, 0, 0, 0.54353);
    -gtk-icon-shadow: 0 -1px rgba(0, 0, 0, 0.54353);
    box-shadow: inset 0 1px rgba(255, 255, 255, 0.4);
    color: white;
    outline-color: rgba(255, 255, 255, 0.3);
    border-color: #215d9c;
    background-image: image(#3180d4);
    box-shadow: inset 0 1px rgba(255, 255, 255, 0);
    text-shadow: none;
    -gtk-icon-shadow: none; }

/* "Press $KEY to exit fullscreen" overlay text */
#fullscreen-popup {
  padding: 12px;
  border-radius: 18px;
  background: rgba(0, 0, 0, 0.65);
  color: white; }
