/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/*
 *  Copyright © 2003 Christian Persch
 *
 *  This file is part of Epiphany.
 *
 *  Epiphany is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Epiphany is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Epiphany.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "config.h"
#include "ephy-zoom.h"

#include <math.h>

float
ephy_zoom_get_changed_zoom_level (float level, int steps)
{
  float new_level = level;

  if (steps == -1) {
    new_level -= 0.25;
    if (new_level < ZOOM_MINIMAL)
      new_level = ZOOM_MINIMAL;
  } else if (steps == 1) {
    new_level += 0.25;
    if (new_level > ZOOM_MAXIMAL)
      new_level = ZOOM_MAXIMAL;
  }

  return new_level;
}
