/* Global CSS for pages */

h1 {
        color: #babdb6;
        text-shadow: 0 1px 0 white;
        margin-bottom: 0;
}

html {
        font-family: -webkit-system-font, Cantarell, sans-serif;
        font-size: 11pt;
        color: #2e3436;
        padding: 20px 20px 0 20px;
        background-color: #f6f6f4;
        background-image: -webkit-gradient(
                linear,
                left top,
                left bottom,
                color-stop(0, #eeeeec),
                color-stop(1, #f6f6f4)
        );
        background-size: 100% 5em;
        background-repeat: no-repeat;
}

/* about: */

.dialog {
    padding: 2em;
    margin: 2em auto;
    text-align: center;
    background-color: #fff;
    border-radius: 1em;
    border: 2px solid #d3d7cf;
    width: 40em;
}

.properties {
    margin: auto;
}

.prop-label {
    color: #999;
}

#about-title {
    color: black;
}

#about-subtitle {
    color: #babdb6;
    text-shadow: 0 1px 0 white;
}

/* about:plugins */

#plugin-table
{
    margin-top: 10px;
    width: 96%;
    text-align: left;
    border-collapse: collapse;
}

#plugin-table th
{
    padding: 6px;
    background: #565051;
    border: 2px solid #565051;
    color: #f6f6f4;
    width: 33%;
}

#plugin-table td
{
    padding: 4px;
    background: #f6f6f4;
    border-bottom: 1px solid #d3d7cf;
    color: #2e3436;
    border-top: 1px solidtransparent;
}

#plugin-table tr:hover td
{
    background: #d3d7cf;
    color: #2e3436;
}

/* about:epiphany */

html.epiphany-html {
    background-color: #3465a4;
    background-image: none;
}

.epiphany-body {
    color: #eeeeec;
    max-width: 40em;
    margin: 2em auto;
}

#ephytext {
    margin-top: 15%;
    font-size: 1.5em;
    text-align: center;
}

#from {
    margin-top: 1em;
    font-size: 1.5em;
    text-align: right;
    font-weight: bold;
}

/* about:memory */

.memory-table caption {
    font-size: 16pt;
    font-weight: bold;
    margin-bottom: 0.9em;
    color: #2e3436;
    text-shadow: 0 1px 0 white;
}

.memory-table {
    margin: 0 12.5% 0.9em 12.5%;
    width: 80%;
    text-align: left;
    border-collapse: collapse;
}

.memory-table th {
    padding: 4px;
    background: #565051;
    border: 2px solid #565051;
    color: #f6f6f4;
}

.memory-table td {
    padding: 2px;
    background: #f6f6f8;
    border-bottom: 1px solid #d3d7cf;
    color: #2e3436;
    border-top: 1px solid transparent;
    width: 16%;
}

.memory-table tr:hover td {
    background: #d3d7cf;
    color: #2e3436;
}


/* about:applications */

.applications-body table { width: 100%; border-collapse: collapse; }

.applications-body table, td { 
	border: 1px solid #d3d7cf;
	border-left: none; border-right: none;
}

.applications-body p { margin-bottom: 30px; }
.applications-body td { padding: 15px;}
.applications-body td.icon { width: 64px; }
.applications-body td.datas { width: 200px; }
.applications-body td.input { width: 64px; }
.applications-body td.date { width: auto; text-align: right; font-size: small; }
.applications-body .appname { font-weight: bold; }
.applications-body .appurl, td.date { color: #babdb6; }
.applications-body input { width: 100%; padding: 8px; }

/* about:incognito */

.incognito-body {
    margin: 0;
    padding: 0;
}

.incognito-body div#mainblock {
    display: table;
    max-width: 40em;
    margin-left: auto;
    margin-right: auto;
    height: 75%;
}

.incognito-body div#mainblock>div {
    margin: 0 10px;
    display: table-cell;
    vertical-align: middle;
    padding: 0 20px 0 160px;
}

.incognito-body h1 {
    color: #888a85;
}

/* about:overview */
#overview {
    padding: 0 5px;
    max-width: 700px;
    margin-left: auto;
    margin-right: auto;
    bottom: 0;
    max-height: 615px;
    margin: auto;
    position: absolute;
    top: 0;
    right: 0;
    left: 0;
    text-align: center;
}

.overview-item {
    width: 182px;
    height: 185px;
    display: inline-block;
    vertical-align: top;
    overflow: hidden;
    margin: 20px 20px 0 20px;
    outline: 0;
    position: relative;
    transition: 0;
    top: 0;
    text-align: center;
}

.overview-title {
    width: 180px;
    height: 40px;
    display: inline-block;
    margin-top: 10px;
    overflow: hidden;
    font-family: -webkit-system-font, Cantarell, sans-serif;
    font-size: 11pt;
    color: #2e3436;
    text-overflow: ellipsis;
    text-align: center;
}

.overview-thumbnail {
    width: 180px;
    height: 135px;
    display: block;
    border: 1px solid #ccc;
    background: url(ephy-resource:///org/gnome/epiphany/missing-thumbnail.png) center no-repeat, linear-gradient(to bottom, #e8e8e8, #fff);
    background-repeat: no-repeat;
    position: relative;
}

.overview-thumbnail:hover,
:focus .overview-thumbnail {
    border: 1px solid #999;
    box-shadow: 0 1px 1px rgba(0,0,0,0.3);
}

.overview-thumbnail:active {
    border: 1px solid #999;
    box-shadow: inset 0 1px 1px rgba(0,0,0,0.3);
}

.overview-close-button {
    -webkit-transition: all 250ms;
    position: absolute;
    top: 5px;
    right: 5px;
    opacity: 0;
    z-index: 5;
    color: #888;
    font-family: -webkit-system-font, Cantarell, sans-serif;
    font-size: 14px;
    width: 13px;
    height: 14px;
    color: #fff;
    padding: 3px 3px 4px 4px;
    text-shadow: 0 1px 0 rgba(0,0,0,.3);
    background-color: rgba(80,80,80,.2);
    border-radius: 50%;
}

.overview-close-button:hover {
    background-color: rgba(0,0,0,.8);
}

.overview-close-button:active {
    text-shadow: none;
    background-color: #000;
}

.overview-item:hover .overview-close-button {
    opacity: 1;
}

.overview-removed.overview-item {
    transition: all 0.75s ease-in-out;
    width: 0;
    margin: 0;
    opacity: 0;
    padding: 0;
}

.overview-removed .overview-close-button {
    display: none;
}

html[dir="rtl"] .overview-close-button {
    left: 20px;
    right: inherit;
}

html[dir="rtl"] .overview-item {
    float: right;
}

.overview-empty {
    display: flex;
    flex-direction: column;
    align-items: center;
    justify-content: center;
    margin-left: auto;
    margin-right: auto;
    height: 80%;
    max-width: 800px;
    opacity: 0.55; /* Adwaita's dim-label */
    -webkit-user-select: none;
    -webkit-user-drag: none;
}

.overview-empty h1 {
    color: black;
    opacity: 0.55; /* Adwaita's dim-label */
    text-shadow: none;
}

.overview-empty img {
    -webkit-user-select: none;
    -webkit-user-drag: none;
    filter: grayscale(100%);
    width: 128px;
}
