/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* 
 *  Copyright © 2005 Christian Persch
 *
 *  This file is part of Epiphany.
 *
 *  Epiphany is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Epiphany is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Epiphany.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

static const char languages[][7] =
{
/* If you're missing your language/locale combination, file a bug at
 * http://bugzilla.gnome.org/enter_bug.cgi?product=epiphany
 */
"ab",
"af",
"af-na",
"af-za",
"am",
"an",
"ar",
"ar-ae",
"ar-bh",
"ar-dj",
"ar-dz",
"ar-eg",
"ar-eh",
"ar-er",
"ar-il",
"ar-iq",
"ar-jo",
"ar-km",
"ar-kw",
"ar-lb",
"ar-ly",
"ar-ma",
"ar-mr",
"ar-om",
"ar-ps",
"ar-qa",
"ar-sa",
"ar-sd",
"ar-sy",
"ar-td",
"ar-tn",
"ar-ye",
"ast",
"as",
"as-bd",
"as-in",
"ay",
"ay-bo",
"ay-pe",
"az",
"az-az",
"az-ge",
"az-ir",
"az-tr",
"bal",
"bal-af",
"bal-ir",
"bal-om",
"be",
"bg",
"bn",
"br",
"bn-bd",
"bn-in",
"bo",
"bo-cn",
"bo-in",
"bo-np",
"bs",
"bs-cs",
"bs-hr",
"ca",
"ce",
"ca-ad",
"ca-es",
"ch",
"co",
"ch-gu",
"ch-mp",
"chr",
"cpe",
"cr",
"cs",
"cv",
"cy",
"da",
"da-dk",
"da-gl",
"de",
"de-at",
"de-be",
"de-ch",
"de-de",
"de-li",
"de-lu",
"de-na",
"dgr",
"dv",
"dz",
"el",
"el-cy",
"el-gr",
"en",
"en-ag",
"en-ai",
"en-as",
"en-au",
"en-aw",
"en-bb",
"en-bm",
"en-bs",
"en-bw",
"en-bz",
"en-ca",
"en-cc",
"en-ck",
"en-cm",
"en-cx",
"en-dm",
"en-er",
"en-fj",
"en-fk",
"en-fm",
"en-gb",
"en-gd",
"en-gh",
"en-gi",
"en-gm",
"en-gy",
"en-hk",
"en-ie",
"en-in",
"en-io",
"en-jm",
"en-ke",
"en-ki",
"en-kn",
"en-ky",
"en-lc",
"en-lr",
"en-ls",
"en-mh",
"en-mp",
"en-ms",
"en-mt",
"en-mw",
"en-na",
"en-nf",
"en-ng",
"en-nr",
"en-nu",
"en-nz",
"en-pg",
"en-ph",
"en-pk",
"en-pn",
"en-pw",
"en-rw",
"en-sb",
"en-sg",
"en-sh",
"en-sl",
"en-sz",
"en-tc",
"en-tk",
"en-to",
"en-tt",
"en-tz",
"en-ug",
"en-um",
"en-us",
"en-vc",
"en-vg",
"en-vi",
"en-za",
"en-zm",
"en-zw",
"eo",
"es",
"es-ar",
"es-bo",
"es-cl",
"es-co",
"es-cr",
"es-cu",
"es-do",
"es-ec",
"es-es",
"es-gi",
"es-gq",
"es-gt",
"es-hn",
"es-mx",
"es-ni",
"es-pa",
"es-pe",
"es-pr",
"es-py",
"es-sv",
"es-us",
"es-uy",
"es-ve",
"et",
"eu",
"fa",
"fa-af",
"fa-ir",
"fi",
"fi-fi",
"fil",
"fi-se",
"fiu",
"fiu-fi",
"fiu-se",
"fj",
"fo",
"fr",
"fr-be",
"fr-bf",
"fr-bi",
"fr-bj",
"fr-ca",
"fr-cd",
"fr-cf",
"fr-cg",
"fr-ch",
"fr-ci",
"fr-cm",
"fr-dj",
"fr-fr",
"fr-ga",
"fr-gf",
"fr-gn",
"fr-gp",
"fr-gq",
"fr-ht",
"fr-it",
"fr-km",
"fr-lu",
"fr-mc",
"fr-mg",
"fr-ml",
"fr-mq",
"fr-mr",
"fr-mu",
"fr-nc",
"fr-ne",
"fr-pf",
"fr-pm",
"fr-re",
"fr-rw",
"fr-sc",
"fr-sn",
"fr-td",
"fr-tf",
"fr-tg",
"fr-us",
"fr-vu",
"fr-wf",
"fr-yt",
"fur",
"fy",
"fy-de",
"fy-nl",
"ga",
"gd",
"gl",
"gn",
"gn-ar",
"gn-bo",
"gu",
"gv",
"ha",
"ha-ne",
"ha-ng",
"haw",
"he",
"hsb",
"hi",
"hi-in",
"ho",
"hr",
"hr-ba",
"hr-hr",
"ht",
"hu",
"hu-cs",
"hu-si",
"hy",
"ia",
"hy-az",
"hy-ge",
"hy-ir",
"id",
"ie",
"is",
"it",
"it-ch",
"it-hr",
"it-it",
"it-si",
"it-sm",
"it-va",
"iu",
"iu-ca",
"iu-gl",
"iu-ru",
"iu-us",
"ja",
"ja-gu",
"ja-jp",
"ja-pw",
"ka",
"kk",
"kl",
"km",
"kn",
"ko",
"kok",
"ko-kp",
"ko-kr",
"ks",
"ku",
"ku-iq",
"ku-ir",
"ku-tr",
"kw",
"ky",
"la",
"lb",
"li",
"ln",
"lo",
"lt",
"lv",
"mai",
"men",
"men-gn",
"men-lr",
"men-sl",
"mg",
"mh",
"mh-mh",
"mh-nr",
"mi",
"mi-ck",
"mi-nz",
"mk",
"ml",
"mn",
"mni",
"mo",
"mr",
"ms",
"ms-bn",
"ms-cc",
"ms-cx",
"ms-my",
"ms-sg",
"mt",
"my",
"na",
"nb",
"nb-no",
"nb-sj",
"nd",
"ne",
"ng",
"niu",
"nl",
"nl-an",
"nl-aw",
"nl-be",
"nl-nl",
"nl-sr",
"nn",
"nn-no",
"nn-sj",
"no",
"no-no",
"no-sj",
"nr",
"nso",
"ny",
"ny-mz",
"ny-zm",
"oc",
"oj",
"om",
"om-et",
"om-ke",
"or",
"os",
"os-ge",
"os-ru",
"pa",
"pa-in",
"pa-pk",
"pau",
"pl",
"ps",
"ps-af",
"ps-pk",
"pt",
"pt-ao",
"pt-br",
"pt-cv",
"pt-gw",
"pt-mo",
"pt-mz",
"pt-pt",
"pt-st",
"pt-tl",
"qu",
"qu-bo",
"qu-pe",
"rm",
"rn",
"ro",
"ru",
"ru-by",
"ru-ge",
"ru-kg",
"ru-kz",
"ru-ru",
"ru-sj",
"ru-ua",
"rw",
"sa",
"sc",
"sd",
"sd-in",
"sd-pk",
"sg",
"si",
"sk",
"sl",
"sm",
"sm-as",
"sm-ws",
"so",
"so-dj",
"so-et",
"so-ke",
"son",
"son-bj",
"son-ml",
"son-ne",
"so-so",
"sq",
"sq-al",
"sq-cs",
"sq-mk",
"sr",
"sr-ba",
"sr-cs",
"ss",
"ss-sz",
"ss-za",
"st",
"st-ls",
"st-za",
"sv",
"sv-ax",
"sv-fi",
"sv-se",
"sw",
"sw-ke",
"sw-tz",
"sw-yt",
"ta",
"ta-in",
"ta-lk",
"ta-sg",
"te",
"tem",
"tet",
"tg",
"th",
"ti",
"ti-er",
"ti-et",
"tig",
"tk",
"tk-af",
"tk-ir",
"tkl",
"tl",
"tl-gu",
"tl-ph",
"tmh",
"tmh-ml",
"tmh-ne",
"tn",
"tn-bw",
"tn-za",
"to",
"tpi",
"tr",
"tr-cy",
"tr-tr",
"ts",
"tt",
"tvl",
"uk",
"ur",
"ur-in",
"ur-pk",
"uz",
"ve",
"vi",
"vo",
"wa",
"wo",
"wo-gm",
"wo-mr",
"wo-sn",
"xh",
"yi",
"xh-ls",
"xh-za",
"zh",
"zh-cn",
"zh-cx",
"zh-hk",
"zh-mo",
"zh-sg",
"zh-tw",
"zu",
"zu-ls",
"zu-za"
};
